import copy
import warnings
from typing import List, Optional

import pandas as pd

warnings.simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

from wise_pizza.plotting import plot_segments, plot_split_segments, plot_waterfall
from wise_pizza.slicer import SliceFinder, SlicerPair
from wise_pizza.utils import diff_dataset, prepare_df


def explain_changes_in_average(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    dims: List[str],
    total_name: str,
    size_name: str,
    min_segments: int = 5,
    max_segments: Optional[int] = None,
    min_depth: int = 1,
    max_depth: int = 2,
    solver: str = "lasso",
    how: str = "totals",
    force_add_up: bool = False,
    constrain_signs: bool = True,
    verbose: int = 0,
):
    """
    Find segments most useful in explaining the difference between the averages of the two datasets
    @param df1: First dataset
    @param df2: Second dataset
    @param dims: List of discrete dimensions
    @param total_name: Name of column that contains totals per segment
    @param size_name: Name of column containing segment sizes
    @param min_segments: Minimum number of segments to find
    @param max_segments: Maximum number of segments to find, defaults to min_segments
    @param min_depth: Minimum number of dimension to constrain in segment definition
    @param max_depth: Maximum number of dimension to constrain in segment definition
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param how: "totals" to only decompose segment totals (ignoring size vs average contribution)
            "split_fits" to separately decompose contribution of size changes and average changes
            "extra_dim" to treat size vs average change contribution as an additional dimension
            "force_dim" like extra_dim, but each segment must contain a Change_from constraint
    @param force_add_up: Force the contributions of chosen segments to add up
    to the difference between dataset totals
    @param constrain_signs: Whether to constrain weights of segments to have the same
    sign as naive segment averages
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @return: A fitted object
    """
    df1 = df1.copy()
    df2 = df2.copy()

    # replace NaN values in numeric columns with zeros
    # replace NaN values in categorical columns with the column name + "_unknown"
    df1 = prepare_df(df1, dims, size_name, total_name)
    df2 = prepare_df(df2, dims, size_name, total_name)

    # rescale sizes and totals, preserving averages (= total/size)
    df1["Norm_weight"] = df1[size_name] / df1[size_name].sum()
    df2["Norm_weight"] = df2[size_name] / df2[size_name].sum()

    df1["Norm_totals"] = df1[total_name] / df1[size_name].sum()
    df2["Norm_totals"] = df2[total_name] / df2[size_name].sum()

    # subtract the initial average from both totals
    avg1 = df1["Norm_totals"].sum()
    df1["Adj_totals"] = df1["Norm_totals"] - avg1 * df1["Norm_weight"]
    df2["Adj_totals"] = df2["Norm_totals"] - avg1 * df2["Norm_weight"]

    # call explain_changes
    sf = explain_changes_in_totals(
        df1,
        df2,
        dims,
        total_name="Adj_totals",
        size_name="Norm_weight",
        min_segments=min_segments,
        max_segments=max_segments,
        min_depth=min_depth,
        max_depth=max_depth,
        solver=solver,
        how=how,
        force_add_up=force_add_up,
        constrain_signs=constrain_signs,
        verbose=verbose,
    )

    if hasattr(sf, "pre_total"):
        sf.pre_total = avg1
        sf.post_total += avg1
        sfs = None
    # Want to put the subtracted avg1 back in, something like
    # for s in sf.segments:
    #     s["naive_avg"] += average
    #     s["total"] += average * s["seg_size"]
    # print(average)
    # sf.reg.intercept_ = average

    # And might want to relabel some plots?
    sf.task = "changes in average"
    return sf


def explain_changes_in_totals(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    dims: List[str],
    total_name: str,
    size_name: str,
    min_segments: int = 5,
    max_segments: Optional[int] = None,
    min_depth: int = 1,
    max_depth: int = 2,
    solver: str = "lasso",
    how: str = "totals",
    force_add_up: bool = False,
    constrain_signs: bool = True,
    verbose: int = 0,
):
    """
    Find segments most useful in explaining the difference between the totals of the two datasets
    @param df1: First dataset
    @param df2: Second dataset
    @param dims: List of discrete dimensions
    @param total_name: Name of column that contains totals per segment
    @param size_name: Name of column containing segment sizes
    @param min_segments: Minimum number of segments to find
    @param max_segments: Maximum number of segments to find, defaults to min_segments
    @param min_depth: Minimum number of dimension to constrain in segment definition
    @param max_depth: Maximum number of dimension to constrain in segment definition
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param how: "totals" to only decompose segment totals (ignoring size vs average contribution)
            "split_fits" to separately decompose contribution of size changes and average changes
            "extra_dim" to treat size vs average change contribution as an additional dimension
            "force_dim" like extra_dim, but each segment must contain a Change_from constraint
    @param force_add_up: Force the contributions of chosen segments to add up
    to the difference between dataset totals
    @param constrain_signs: Whether to constrain weights of segments to have the same
    sign as naive segment averages
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @return: A fitted object
    """

    assert how in ["totals", "extra_dim", "split_fits", "force_dim"]
    split_deltas = not (how == "totals")
    return_multiple = how == "split_fits"
    final_size = df2[size_name].sum()

    df1 = df1.copy()
    df2 = df2.copy()

    # replace NaN values in numeric columns with zeros
    # replace NaN values in categorical columns with the column name + "_unknown"
    df1 = prepare_df(df1, dims, size_name, total_name)
    df2 = prepare_df(df2, dims, size_name, total_name)

    my_diff = diff_dataset(
        df1,
        df2,
        dims,
        total_name,
        size_name,
        split_deltas=split_deltas,
        return_multiple=return_multiple,
    )

    if how == "split_fits":
        df_size, df_avg = my_diff
        sf_size = explain_levels(
            df=df_size.data,
            dims=dims,
            total_name=df_size.segment_total,
            size_name=df_size.segment_size,
            min_depth=min_depth,
            max_depth=max_depth,
            min_segments=min_segments,
            solver=solver,
            force_add_up=force_add_up,
            constrain_signs=constrain_signs,
            verbose=verbose,
        )

        sf_avg = explain_levels(
            df=df_avg.data,
            dims=dims,
            total_name=df_avg.segment_total,
            size_name=df_avg.segment_size,
            min_depth=min_depth,
            max_depth=max_depth,
            min_segments=min_segments,
            solver=solver,
            force_add_up=force_add_up,
            constrain_signs=constrain_signs,
            verbose=verbose,
        )

        sf_size.final_size = final_size
        sf_avg.final_size = final_size
        sp = SlicerPair(sf_size, sf_avg)
        sp.plot = (
            lambda plot_is_static=False, width=2000, height=500: plot_split_segments(
                sp.s1,
                sp.s2,
                plot_is_static=plot_is_static,
                width=width,
                height=height,
            )
        )
        return sp

    else:
        sf = SliceFinder()

        sf.fit(
            my_diff.data[my_diff.dimensions],
            my_diff.data[my_diff.segment_total],
            weights=my_diff.data[my_diff.segment_size],
            min_depth=min_depth,
            max_depth=max_depth,
            min_segments=min_segments,
            max_segments=max_segments,
            solver=solver,
            force_dim="Change from" if how == "force_dim" else None,
            force_add_up=force_add_up,
            constrain_signs=constrain_signs,
            verbose=verbose,
        )

        sf.pre_total = df1[total_name].sum()
        sf.post_total = df2[total_name].sum()

        sf.plot = lambda plot_is_static=False, width=1000, height=1000: plot_waterfall(
            sf, plot_is_static=plot_is_static, width=width, height=height
        )
        sf.task = "changes in totals"
        return sf


def explain_levels(
    df: pd.DataFrame,
    dims: List[str],
    total_name: str,
    size_name: Optional[str] = None,
    min_segments: int = 10,
    max_segments: int = None,
    min_depth: int = 1,
    max_depth: int = 2,
    solver="lasso",
    verbose=0,
    force_add_up: bool = False,
    constrain_signs: bool = True,
):
    """
    Find segments whose average is most different from the global one
    @param df: Dataset
    @param dims: List of discrete dimensions
    @param total_name: Name of column that contains totals per segment
    @param size_name: Name of column containing segment sizes
    @param min_segments: Minimum number of segments to find
    @param max_segments: Maximum number of segments to find, defaults to min_segments
    @param min_depth: Minimum number of dimension to constrain in segment definition
    @param max_depth: Maximum number of dimension to constrain in segment definition
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @param force_add_up: Force the contributions of chosen segments to add up to zero
    @param constrain_signs: Whether to constrain weights of segments to have the same
    sign as naive segment averages
    @return: A fitted object
    """
    df = copy.copy(df)

    # replace NaN values in numeric columns with zeros
    # replace NaN values in categorical columns with the column name + "_unknown"
    df = prepare_df(df, dims, size_name, total_name)

    if size_name is None:
        size_name = "size"
        df[size_name] = 1.0

    # we want to look for deviations from average value
    average = df[total_name].sum() / df[size_name].sum()
    df["_target"] = df[total_name] - df[size_name] * average

    sf = SliceFinder()
    sf.fit(
        df[dims],
        df["_target"],
        weights=None if size_name is None else df[size_name],
        min_segments=min_segments,
        max_segments=max_segments,
        min_depth=min_depth,
        max_depth=max_depth,
        solver=solver,
        verbose=verbose,
        force_add_up=force_add_up,
        constrain_signs=constrain_signs,
    )

    for s in sf.segments:
        s["naive_avg"] += average
        s["total"] += average * s["seg_size"]
    # print(average)
    sf.reg.intercept_ = average
    sf.plot = lambda plot_is_static=False, width=2000, height=500: plot_segments(
        sf, plot_is_static=plot_is_static, width=width, height=height
    )
    sf.task = "levels"
    return sf



import itertools
from typing import Optional, List, Dict

import numpy as np
import scipy
from tqdm import tqdm
import pandas as pd
from scipy.sparse import csc_matrix, hstack


def join_to_sparse(dim_df: pd.DataFrame, dim_name: str, verbose=0):
    values = sorted(dim_df[dim_name].unique())

    # create an "eye" dataframe
    ext_df = pd.DataFrame(data=np.eye(len(values)), columns=values)
    ext_df[dim_name] = values

    join_df = pd.merge(dim_df, ext_df, on=[dim_name])
    join_df = join_df.sort_values(list(dim_df.columns))
    vals = csc_matrix(join_df[values].values)
    if verbose > 0:
        print(values, vals.shape)
    return vals, values


def segment_defs(dim_df: pd.DataFrame, used_dims, verbose=0) -> List[Dict[str, str]]:
    col_defs = []
    this_df = dim_df[used_dims].drop_duplicates().reset_index(drop=True)
    # create an "eye" dataframe on the unique reduced dimensions
    for i, vals in enumerate(this_df.itertuples(index=False)):
        col_defs.append(dict(zip(used_dims, vals)))

    if verbose > 0:
        print(used_dims, len(col_defs))
    return col_defs


def construct_dummies(
    segment_defs: List[Dict[str, str]], cache: Dict[str, Dict[str, np.ndarray]]
) -> scipy.sparse.csc_matrix:
    dummies = []
    for sgdf in segment_defs:
        tmp = None
        for k, v in sgdf.items():
            if tmp is None:
                tmp = cache[k][v]
            else:
                tmp = tmp.multiply(cache[k][v])
        dummies.append(tmp)
    return hstack(dummies)


# This approach was way slower than the join one; keeping it here for reference :)
# def join_to_sparse(dim_df, this_df, chunk_size=100, verbose=0):
#     mats = []
#     tuples = []
#     col_defs = []
#
#     these_dims = list(this_df.columns)
#     # create an "eye" dataframe on the unique reduced dimensions
#     for i, vals in enumerate(this_df.itertuples(index=False)):
#         tuples.append(vals)
#         col_defs.append({k: v for k, v in zip(these_dims, vals)})
#
#     # join it against the real thing, one chunk at a time
#     for i in range(0, len(tuples), chunk_size):
#         this_df = dim_df.copy()
#         these_cols = []
#         for tpl in tuples[i : min(i + chunk_size, len(tuples))]:
#             col_name = "_".join(map(str, tpl))
#             these_cols.append(col_name)
#             this_df[col_name] = 0.0
#
#             for i, (col, value) in enumerate(zip(these_dims, tpl)):
#                 if i == 0:
#                     filter = this_df[col] == value
#                 else:
#                     filter = filter & (this_df[col] == value)
#
#             this_df[filter] = 1.0
#
#         vals = csc_matrix(this_df[these_cols].values)
#         del this_df
#
#         mats.append(vals)
#         # print(vals.shape)
#     if len(mats) > 1:
#         out = hstack(mats)
#     else:
#         out = mats[0]
#     if verbose > 0:
#         print(these_dims, out.shape)
#     return out, col_defs


def sparse_dummy_matrix(
    dim_df: pd.DataFrame,
    min_depth: int = 1,
    max_depth: int = 2,
    verbose=0,
    force_dim: Optional[str] = None,
):
    # generate a sparse dummy matrix based on all the combinations
    # TODO: do a  nested sparse regression fit to form groups of dim values, pos, neg, null
    # TODO: first calculate the matrix size, scale down max_depth if matrix too big
    if force_dim is None:
        dims = list(dim_df.columns)
    else:
        assert force_dim in dim_df.columns
        dims = [c for c in dim_df.columns if c != force_dim]

    # drop dimensions with only one value, for clarity
    dims = [d for d in dims if len(dim_df[d].unique()) > 1]

    defs = []
    mats = []
    dims_range_min = min(len(dims), max(1, min_depth))
    dims_range_max = min(len(dims) + 1, max_depth + 1)
    dims_range = range(dims_range_min, dims_range_max)

    # first pass: generate single-dim dummies
    dummy_cache = {}
    for d in dim_df.columns:
        this_mat, these_defs = join_to_sparse(dim_df, d, verbose=verbose)
        dummy_cache[d] = {this_def: this_mat[:, i : i + 1] for i, this_def in enumerate(these_defs)}

    for num_dims in tqdm(dims_range) if verbose else dims_range:
        for these_dims in itertools.combinations(dims, num_dims):
            if num_dims == 1 and these_dims[0] == "Change from":
                continue
            if force_dim is None:
                used_dims = list(these_dims)
            else:
                used_dims = [force_dim] + list(these_dims)

            these_defs = segment_defs(dim_df, used_dims, verbose=verbose)
            this_mat = construct_dummies(these_defs, dummy_cache)
            mats.append(this_mat)
            defs += these_defs
    mat = hstack(mats)
    return mat, defs



from typing import Optional

import plotly.graph_objects as go
import plotly.io as pio
from plotly.io import to_image
from plotly.subplots import make_subplots

from wise_pizza.slicer import SliceFinder

pio.templates.default = "plotly_white"

import numpy as np
import pandas as pd
from IPython.display import Image, display


def plot_split_segments(
    sf_size: SliceFinder,
    sf_avg: SliceFinder,
    plot_is_static: bool = False,
    width: int = 2000,
    height: int = 500,
):
    """
    Plot split segments for explain_changes: split_fits
    @param sf_size: SliceFinder from sizes
    @param sf_avg: SliceFinder from averages
    @param denominator: denominator of the value
    @param plot_is_static: static (True) or dynamic (False) plotly result
    @param width: parameter to modify the final width of the plot
    @param height: parameter to modify the final height of the plot
    """
    size_data = pd.DataFrame(sf_size.segments, index=np.array(sf_size.segment_labels))
    avg_data = pd.DataFrame(sf_avg.segments, index=np.array(sf_avg.segment_labels))

    size_average = sf_size.reg.intercept_
    avg_average = sf_avg.reg.intercept_

    trace1 = go.Bar(
        x=size_data["impact"],
        y=size_data.index,
        orientation="h",
        name="Impact of size chgs",
        marker_color="#a0e1e1",
    )

    trace2 = go.Bar(
        x=size_data["naive_avg"],
        y=size_data.index,
        orientation="h",
        name="Diff in sizes",
        marker_color="#ff685f",
    )

    trace3 = go.Bar(
        x=avg_data["impact"],
        y=avg_data.index,
        orientation="h",
        name="Impact of average chgs",
        marker_color="#9fe870",
    )

    trace4 = go.Bar(
        x=avg_data["naive_avg"],
        y=avg_data.index,
        orientation="h",
        name="Diff in averages",
        marker_color="#ffc091",
    )

    fig = make_subplots(
        rows=2,
        cols=2,
        shared_yaxes=True,
        subplot_titles=[
            "Impact of size chgs on overall total",
            "Normalised difference in sizes",
            "Impact of average chgs on overall total",
            "Normalised difference in averages",
        ],
    )

    fig.add_trace(trace1, 1, 1)
    fig.add_trace(trace2, 1, 2)
    fig.add_trace(trace3, 2, 1)
    fig.add_trace(trace4, 2, 2)

    fig.update_layout(
        height=height + len(size_data.index) * 30,
        width=width + len(size_data.index) * 30,
    )

    fig.add_vline(
        x=size_average,
        line_width=3,
        line_dash="dash",
        line_color="red",
        col=2,
        row=1,
        annotation_text="Global average",
    )

    fig.add_vline(
        x=avg_average,
        line_width=3,
        line_dash="dash",
        line_color="red",
        col=2,
        row=2,
        annotation_text="Global average",
    )

    for i in range(1, 3):
        fig.update_yaxes(autorange="reversed", row=i)
    if plot_is_static:
        # Convert the figure to a static image
        image_bytes = to_image(fig, format="png", scale=2)

        # Display the static image in the Jupyter notebook
        return Image(
            image_bytes,
            height=height + len(size_data.index) * 30,
            width=width + len(size_data.index) * 30,
        )
    else:
        fig.show()


def plot_segments(
    sf: SliceFinder,
    plot_is_static: bool = False,
    width: int = 2000,
    height: int = 500,
):
    """
    Plot segments for explain_levels
    @param sf: SliceFinder
    @param plot_is_static: static (True) or dynamic (False) plotly result
    @param width: parameter to modify the final width of the plot
    @param height: parameter to modify the final height of the plot
    """

    seg_data = pd.DataFrame(sf.segments)
    seg_data["avg"] = seg_data["coef"] + sf.reg.intercept_
    seg_data.index = np.array(sf.segment_labels)

    average = sf.reg.intercept_

    trace1 = go.Bar(
        x=seg_data["impact"],
        y=sf.segment_labels,
        orientation="h",
        name="total",
        marker_color="#a0e1e1",
    )

    trace2 = go.Bar(
        x=seg_data["naive_avg"],
        y=sf.segment_labels,
        orientation="h",
        name="averages",
        marker_color="#ff685f",
    )

    trace3 = go.Bar(
        x=seg_data["seg_size"],
        y=sf.segment_labels,
        orientation="h",
        name="sizes",
        marker_color="#9fe870",
    )

    fig = make_subplots(
        rows=1,
        cols=3,
        shared_yaxes=True,
        subplot_titles=[
            "Impact on overall total",
            "Simple segment averages",
            "Segment sizes",
        ],
    )

    fig.add_trace(trace1, 1, 1)
    fig.add_trace(trace2, 1, 2)
    fig.add_trace(trace3, 1, 3)

    fig.update_layout(
        height=height + len(sf.segment_labels) * 30,
        width=width + len(sf.segment_labels) * 30,
        yaxis=dict(autorange="reversed"),
    )
    fig.add_vline(
        x=average,
        line_width=3,
        line_dash="dash",
        line_color="red",
        col=2,
        annotation_text="Global average",
    )

    if plot_is_static:
        # Convert the figure to a static image
        image_bytes = to_image(fig, format="png", scale=2)

        # Display the static image in the Jupyter notebook
        return Image(
            image_bytes,
            height=height + len(sf.segment_labels) * 30,
            width=width + len(sf.segment_labels) * 30,
        )
    else:
        fig.show()


def waterfall_args(sf: SliceFinder):
    """
    Waterfall plot arguments
    @param sf: SliceFinder
    """
    segs = [s["impact"] for s in sf.segments]
    other = sf.post_total - (sf.pre_total + sum(segs))

    def rel_error(x, y):
        return abs(x / y - 1.0)

    assert rel_error(sf.pre_total + sum(segs) + other, sf.post_total) < 1e-3

    return {
        "orientation": "v",
        "measure": ["absolute"] + ["relative"] * (len(sf.segments) + 1) + ["total"],
        "x": ["Old total"] + sf.segment_labels + ["Other", "New total"],
        "textposition": "outside",
        # text = ["+60", "+80", "", "-40", "-20", "Total"],
        "y": [sf.pre_total] + segs + [other, sf.post_total],
        "connector": {"line": {"color": "rgb(63, 63, 63)"}},
        "increasing": {"marker": {"color": "#9fe870"}},
        "decreasing": {"marker": {"color": "#ff685f"}},
        "totals": {"marker": {"color": "#a0e1e1"}},
    }


def waterfall_layout_args(sf: SliceFinder, width: int = 1000, height: int = 1000):
    """
    Waterfall plot layout arguments
    @param sf: SliceFinder
    @param width: parameter to modify the final width of the plot
    @param height: parameter to modify the final height of the plot
    """
    range1 = sum([s["impact"] for s in sf.segments if s["impact"] > 0])
    range2 = abs(sum([s["impact"] for s in sf.segments if s["impact"] < 0]))
    range = max(range1, range2)
    tickangle = 30
    tickfront_size = 8
    return {
        "yaxis_range": [
            min(
                sf.pre_total - range / 4,
                sf.pre_total + range1 - range2 - 0.2 * range,
            ),
            sf.pre_total + range1 + 0.2 * range,
        ],
        "xaxis": {
            "tickangle": tickangle,  # set x-axis tick angle to 30 degrees (horizontal orientation)
            "tickfont": {"size": tickfront_size},  # set font size of x-axis tick labels
        },
        "width": width,
        "height": height,
    }


def plot_waterfall(
    sf: SliceFinder, plot_is_static: bool = False, width: int = 1000, height: int = 1000
):
    """
    Plot waterfall and Bar for explain_changes
    @param sf: SliceFinder
    @param plot_is_static: static (True) or dynamic (False) plotly result
    @param width: parameter to modify the final width of the plot
    @param height: parameter to modify the final height of the plot
    """
    assert hasattr(sf, "pre_total"), "Please call fit_change first"
    data = pd.DataFrame(sf.segments, index=np.array(sf.segment_labels))
    trace1 = go.Waterfall(name="Segments waterfall", **waterfall_args(sf))

    trace2 = go.Bar(
        x=data["naive_avg"],
        y=data.index,
        orientation="h",
        name="Diff in averages",
        marker_color="#ff685f",
    )

    fig = go.Figure()
    fig2 = go.Figure()

    fig.add_trace(trace1)
    fig.update_layout(title="Segments contributing most to the change")
    fig2.add_trace(trace2)
    fig2["layout"]["yaxis"].update(autorange="reversed")
    fig2.update_layout(title="Normalised difference in average")

    fig2.update_layout(width=width, height=height)

    fig.update_layout(
        title="Segments contributing most to the change",
        #         showlegend = True,
        **waterfall_layout_args(sf, width, height)
    )

    if plot_is_static:
        # Convert the figure to a static image
        image_bytes = to_image(fig, format="png", scale=2)
        image_bytes2 = to_image(fig2, format="png", scale=2)

        # Display the static image in the Jupyter notebook
        display(Image(image_bytes, width=width, height=height))
        display(Image(image_bytes2, width=width, height=height))
    else:
        fig.show()
        fig2.show()



import json
from typing import Optional, Union, List, Dict

import numpy as np
import pandas as pd
from scipy.sparse import csc_matrix, diags

from wise_pizza.find_alpha import clean_up_min_max, find_alpha
from wise_pizza.make_matrix import sparse_dummy_matrix


def _summary(obj) -> str:
    out = {"task": obj.task, "segments": obj.segments}
    return json.dumps(out)


class SliceFinder:
    """
    SliceFinder class to find unusual slices
    """

    def __init__(self, **kwargs):
        self.kwargs = kwargs
        self.min_depth = kwargs.get("min_depth", None)
        self.max_depth = kwargs.get("max_depth", None)
        self.dims = None
        self.X = None
        self.col_defs = None
        self.reg = None
        self.nonzeros = None
        self.weights = None
        self.verbose = 0
        self.task = ""

    def _init_mat(
        self,
        dim_df: pd.DataFrame,
        min_depth: int,
        max_depth: int,
        max_cols: int = 300,
        force_dim: Optional[str] = None,
    ):
        """
        Function to initialize sparse matrix
        @param dim_df: Dataset with dimensions
        @param min_depth: Minimum number of dimension to constrain in segment definition
        @param max_depth: Maximum number of dimension to constrain in segment definition
        @param max_cols: Maxumum number of segments to consider
        @param force_dim: To add dim
        @return:
        """
        self.X, self.col_defs = sparse_dummy_matrix(
            dim_df,
            min_depth=min_depth,
            max_depth=max_depth,
            verbose=self.verbose,
            force_dim=force_dim,
        )

        # try naive pre-filter
        if self.X.shape[1] > max_cols:
            chunk_size = int(max_cols / 2)
            # TODO: filter by t-values instead of absolute discrepancies
            seg_wgt = self.X.T @ self.weights
            seg_avg = (self.X.T @ self.totals) / seg_wgt
            avg = self.totals.sum() / self.weights.sum()

            inds = []

            delta = seg_avg - avg
            unusual = np.argsort(np.abs(delta))
            inds += list(unusual[-chunk_size:])

            delta2 = delta * np.sqrt(seg_wgt)
            unusual2 = np.argsort(np.abs(delta2))
            inds += list(unusual2[-chunk_size:])

            delta3 = delta * seg_wgt
            unusual3 = np.argsort(np.abs(delta3))
            inds += list(unusual3[-chunk_size:])

            best = np.array(list(set(inds)))

            self.X = self.X[:, best]
            self.col_defs = [self.col_defs[i] for i in best]
            # end naive pre-filter

        self.min_depth = min_depth
        self.max_depth = max_depth
        self.dims = list(dim_df.columns)

    def fit(
        self,
        dim_df: pd.DataFrame,
        totals: pd.Series,
        weights: pd.Series = None,
        min_segments: int = 10,
        max_segments: int = None,
        min_depth: int = 1,
        max_depth: int = 3,
        solver: str = "lp",
        verbose: Union[bool, None] = None,
        force_dim: Optional[str] = None,
        force_add_up: bool = False,
        constrain_signs: bool = True,
    ):
        """
        Function to fit slicer and find segments
        @param dim_df: Dataset with dimensions
        @param totals: Column with totals
        @param weights: Column with sizes
        @param min_segments: Minimum number of segments to find
        @param max_segments: Maximum number of segments to find, defaults to min_segments
        @param min_depth: Minimum number of dimension to constrain in segment definition
        @param max_depth: Maximum number of dimension to constrain in segment definition
        @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
        @param verbose: If set to a truish value, lots of debug info is printed to console
        @param force_dim: To add dim
        @param force_add_up: To force add up
        @param constrain_signs: To constrain signs
        """
        min_segments, max_segments = clean_up_min_max(min_segments, max_segments)
        if verbose is not None:
            self.verbose = verbose

        totals = np.array(totals).astype(np.float64)

        if weights is None:
            weights = np.ones_like(totals)
        else:
            weights = np.array(weights).astype(np.float64)
        assert min(weights) >= 0
        assert np.sum(np.abs(totals[weights == 0])) == 0

        dims = list(dim_df.columns)
        # sort the dataframe by dimension values,
        # making sure the other vectors stay aligned
        dim_df = dim_df.reset_index(drop=True)
        dim_df["totals"] = totals
        dim_df["weights"] = weights
        dim_df = dim_df.sort_values(dims)
        dim_df = dim_df[dim_df["weights"] != 0]
        self.weights = dim_df["weights"].values
        self.totals = dim_df["totals"].values
        dim_df = dim_df[dims]

        # lazy calculation of the dummy matrix (calculation can be very slow)
        if (
            list(dim_df.columns) != self.dims
            or max_depth != self.max_depth
            or self.X is not None
            and len(dim_df) != self.X.shape[1]
        ):
            self._init_mat(dim_df, min_depth, max_depth, force_dim=force_dim)
        Xw = csc_matrix(diags(self.weights) @ self.X)

        self.reg, self.nonzeros = find_alpha(
            Xw,
            self.totals,
            max_nonzeros=max_segments,
            solver=solver,
            min_nonzeros=min_segments,
            verbose=self.verbose,
            adding_up_regularizer=force_add_up,
            constrain_signs=constrain_signs,
        )

        self.segments = [{"segment": self.col_defs[i]} for i in self.nonzeros]
        wgts = np.array(Xw[:, self.nonzeros].sum(axis=0))[0]

        for i, s in enumerate(self.segments):
            s["coef"] = self.reg.coef_[i]
            s["impact"] = s["coef"] * wgts[i]
            s["avg_impact"] = s["impact"] / sum(wgts)
            s["total"] = (self.totals * self.X[:, self.nonzeros[i]]).sum()
            s["seg_size"] = wgts[i]
            s["naive_avg"] = s["total"] / wgts[i]

        self.segments = self.order_segments(self.segments)

        # In some cases (mostly in a/b exps we have a situation where there is no any diff in totals/sizes)
        if len(self.segments) == 0:
            self.segments.append(
                {
                    "segment": {"No unusual segments": "No unusual segments"},
                    "coef": 0,
                    "impact": 0,
                    "avg_impact": 0,
                    "total": 0,
                    "seg_size": 0,
                    "naive_avg": 0,
                }
            )

    @staticmethod
    def order_segments(segments: List[Dict[str, any]]):
        pos_seg = [s for s in segments if s["impact"] > 0]
        neg_seg = [s for s in segments if s["impact"] < 0]

        return sorted(pos_seg, key=lambda x: abs(x["impact"]), reverse=True) + sorted(
            neg_seg, key=lambda x: abs(x["impact"]), reverse=True
        )

    @staticmethod
    def segment_to_str(segment: Dict[str, any]):
        s = {
            k: v
            for k, v in segment.items()
            if k not in ["coef", "impact", "avg_impact"]
        }
        return str(s)

    @property
    def segment_labels(self):
        return [self.segment_to_str(s["segment"]) for s in self.segments]

    def summary(self):
        return _summary(self)


class SlicerPair:
    def __init__(self, s1: SliceFinder, s2: SliceFinder):
        self.s1 = s1
        self.s2 = s2
        self.task = ""

    def summary(self):
        return _summary(self)



from sklearn.linear_model import Lasso
import numpy as np
import copy
from scipy.sparse import issparse
from scipy.optimize import linprog


def solve_lasso(
    X,
    y,
    alpha,
    constrain_signs=False,
    verbose=None,
    drop_last_row=True,
):
    """
    Lasso-based finder of unusual segments
    @param X: Matrix describing the segments
    @param y: target vector
    @param alpha: regularizer weight
    @param constrain_signs: Whether to constrain weights of segments to have the same
    sign as naive segment averages
    @param verbose: whether to print to output
    @param drop_last_row: Should be set to True if the last row of X
    is used to coerce adding-up of segment impacts to the total y average
    @return: Object returned by Lasso or LinearRegression
    """
    if issparse(X):
        X = X.todense()

    if alpha <= 0:
        alpha = 1e-6

    lasso_args = {
        "max_iter": int(1e5),
        "fit_intercept": False,
        "selection": "random",
        "positive": constrain_signs,  # forces the coefficients to be positive if True,
        "random_state": 42,
    }

    if constrain_signs:
        # make sure the segment coeffs have the same sign as the naive segment impact
        # last row of X is totals normalization,
        X = np.copy(X)
        if drop_last_row:
            Xy = X[:-1].T.dot(y[:-1])
        else:
            Xy = X.T.dot(y)
        neg_idxs = np.where(Xy < 0)[0]  # finding columns with negative
        X[:, neg_idxs] *= -1  # flip signs of X with negative averages

    lasso = Lasso(alpha=alpha, **lasso_args)
    lasso.fit(X, y)
    if constrain_signs:
        lasso.coef_[neg_idxs] *= -1  # flip the signs of the corresponding coefficients
    if verbose:
        nonzeros = np.nonzero(lasso.coef_)[0]
        print(alpha, nonzeros, lasso.coef_[nonzeros])
    return lasso


def solve_lp(X, y, alpha, constrain_signs=False, verbose=None, drop_last_row=True):
    """
    LP-based finder of unusual segments
    @param X: Matrix describing the segments
    @param y: target vector
    @param alpha: regularizer weight
    @param constrain_signs: Whether to constrain weights of segments to have the same
    sign as naive segment averages
    @param verbose: whether to print to output
    @param drop_last_row: Should be set to True if the last row of X
    is used to coerce adding-up of segment impacts to the total y average
    @return: Object returned by linprog, with coef_ and intercept_ attached for convenience
    """
    # TODO: try neater handling of sparse matrices
    if issparse(X):
        X = X.todense()

    """
    The math is as follows: 
    What we want is 
    | X @ a - y |_1 + \alpha | a |_1 -> min
    We use the usual trick to convert 1-norm into an LP problem,
    a := ap - an, 
    X@(ap - an) = sp - sn, 
    so x = [ap, an, sp, sn], x >=0
    and the objective is then 
    sp + sn + \alpha ( ap + an) -> min`
    """

    X_ = copy.copy(X)
    m = X.shape[1]

    c = np.ones(2 * m + 2 * len(y))
    A_eq = np.concatenate([X_, -X_, -np.eye(len(y)), np.eye(len(y))], axis=1)
    bounds = [(0.0, None) for i in range(len(c))]

    if constrain_signs:
        # make sure the segment coeffs have the same sign as the naive segment impact
        # last row of X is totals normalization,
        if drop_last_row:
            Xys = np.sign(X[:-1].T.dot(y[:-1]))
        else:
            Xys = np.sign(X.T.dot(y))

        Xys = np.array(Xys).reshape(-1)
        for i in range(m):
            if Xys[i] > 0:
                bounds[m + i] = (0, 0)
            else:
                bounds[i] = (0, 0)

    c[: 2 * m] *= alpha

    assert len(c) == A_eq.shape[1]
    x = linprog(c, A_eq=A_eq, b_eq=y, bounds=bounds)

    # for compatibility with the sklearn.linear models
    x.coef_ = x.x[:m] - x.x[m : 2 * m]  # * Xys
    x.intercept_ = x.x[2 * m]
    if verbose:
        nonzeros = np.nonzero(x.coef_)[0]
        print(alpha, nonzeros, x.coef_[nonzeros], "status:", x.status)
    return x



from typing import List, Optional

import numpy as np
import pandas as pd

from wise_pizza.segment_data import SegmentData


def diff_dataset(
    df_pre: pd.DataFrame,
    df_post: pd.DataFrame,
    dims: List[str],
    totals: str,
    weights: Optional[str] = None,
    split_deltas: bool = False,
    return_multiple: bool = False,
):
    """
    Create a single dataset that breaks down the change in totals
    to pct change in average and segment size
    @param df_pre: first dataset
    @param df_post: second dataset
    @param dims: List of discrete dimensions
    @param totals: Name of column that contains totals per segment
    @param weights: Name of column containing segment sizes
    @param split_deltas: for calculating differences in sizes and totals
    @param return_multiple: to return multiple prepared datasets
    @return: prepared dataset
    """

    if weights is None:
        split_deltas = False
        weights = "dummy_weight"
        df_pre[weights] = 1.0
        df_post[weights] = 1.0

    avg = "average"

    cols = dims + [totals, weights]

    df_pre = df_pre[cols].groupby(dims, as_index=False).sum()
    df_post = df_post[cols].groupby(dims, as_index=False).sum()

    # create some ratios for pro-rating baselimes
    totals_ratio = df_post[totals].sum() / df_pre[totals].sum()
    weight_ratio = df_post[weights].sum() / df_pre[weights].sum()
    avg_ratio = totals_ratio / weight_ratio

    combined = pd.merge(df_pre, df_post, on=dims, how="outer")

    def rel_error(x, y):
        return abs(x - y) / (abs(x) + abs(y) + 1.0)

    eps = 1e-6
    # use for testing
    # assert rel_error(combined[weights + "_x"].sum(), df_pre[weights].sum()) < eps
    # assert rel_error(combined[weights + "_y"].sum(), df_post[weights].sum()) < eps
    # assert rel_error(combined[totals + "_x"].sum(), df_pre[totals].sum()) < eps
    # assert rel_error(combined[totals + "_y"].sum(), df_post[totals].sum()) < eps

    # clean up the NaNs
    combined[avg + "_x"] = (combined[totals + "_x"] / combined[weights + "_x"]).fillna(
        df_pre[totals].sum() / df_pre[weights].sum()
    )
    combined[avg + "_y"] = (combined[totals + "_y"] / combined[weights + "_y"]).fillna(
        df_post[totals].sum() / df_post[weights].sum()
    )
    combined[weights + "_x"] = combined[weights + "_x"].fillna(0.0)
    combined[weights + "_y"] = combined[weights + "_y"].fillna(0.0)
    combined[totals + "_x"] = combined[totals + "_x"].fillna(0.0)
    combined[totals + "_y"] = combined[totals + "_y"].fillna(0.0)

    combined[weights] = 0.5 * (combined[weights + "_y"] + combined[weights + "_x"])

    # # Baseline prediction for next period is just pro-rated from the totals
    # # Why did I want that in the first place? :)
    # combined["baseline_" + weights] = combined[weights + "_x"] * weight_ratio
    #
    # combined["baseline_" + avg] = combined[avg + "_x"] * avg_ratio
    # combined["baseline_" + totals] = (
    #     combined["baseline_" + weights] * combined["baseline_" + avg]
    # )
    #
    # assert (
    #     rel_error(combined["baseline_" + weights].sum(), combined[weights + "_y"].sum())
    #     < eps
    # )
    #
    # assert (
    #     rel_error(combined["baseline_" + totals].sum(), combined[totals + "_y"].sum())
    #     < eps
    # )

    if split_deltas:  # TODO: add multiplier-based baselines for the diffs
        combined["dweights"] = combined[weights + "_y"] - combined[weights + "_x"]
        combined["davg"] = combined[avg + "_y"] - combined[avg + "_x"]

        # a1 * w1 - a0 * w0 = a1 ( w1 - w0) + w0 ( a1 - a0)
        W0 = combined[weights + "_x"].sum()
        A1 = combined[totals + "_y"].sum()

        # scale the deltas so w and avg deltas are on same scale
        change_from_w = combined["dweights"] * A1
        w_weights = combined[avg + "_y"] / A1

        change_from_avg = combined["davg"] * W0
        avg_weights = combined[weights + "_x"] / W0

        re_change = change_from_w * w_weights + change_from_avg * avg_weights
        combined["dtotals"] = combined[totals + "_y"] - combined[totals + "_x"]

        # assert (combined["dtotals"] - re_change).abs().max() < 0.1

        combined["Change in totals"] = change_from_w * w_weights
        combined[weights] = 1.0  # w_weights

        c2 = combined.copy()
        c2["Change in totals"] = change_from_avg * avg_weights
        c2[weights] = 1.0  # avg_weights

        if return_multiple:
            sd_size = SegmentData(
                combined.rename(
                    columns={"Change in totals": "Change from segment size"}
                ),
                dimensions=dims,
                segment_total="Change from segment size",
                segment_size=weights,
            )
            sd_avg = SegmentData(
                c2.rename(columns={"Change in totals": "Change from segment average"}),
                dimensions=dims,
                segment_total="Change from segment average",
                segment_size=weights,
            )
            return sd_size, sd_avg

        else:
            combined["Change from"] = "Segment size"
            c2["Change from"] = "Segment average"

            df = pd.concat([combined, c2])[
                dims + [weights, "Change in totals", "Change from"]
            ]
            df_change_in_totals = np.array(df["Change in totals"], dtype=np.longdouble)
            combined_dtotals = np.array(combined["dtotals"], dtype=np.longdouble)
            df_change_in_totals_sum = np.nansum(df_change_in_totals)
            combined_dtotals_sum = np.nansum(combined_dtotals)
            # if combined_dtotals_sum > 1e-31:
            #     assert rel_error(df_change_in_totals_sum, combined_dtotals_sum) < eps

            return SegmentData(
                df,
                dimensions=dims + ["Change from"],
                segment_total="Change in totals",
                segment_size=weights,
            )

    else:
        combined["Change in totals"] = combined[totals + "_y"] - combined[totals + "_x"]

        # TODO: why does taking prev period as baseline make the fit so slow?
        # baseline hypothesis is proportional increase from first period
        combined[weights] = 1.0  # combined[totals + "_x"]
        combined[weights] = np.maximum(1.0, combined[weights])
        cols = (
            dims
            + ["Change in totals", totals + "_x", totals + "_y"]
            + [c for c in combined.columns if "baseline" in c]
        )

        return SegmentData(
            combined[cols + [weights]],
            dimensions=dims,
            segment_total="Change in totals",
            segment_size=weights,
        )


def prepare_df(
    df: pd.DataFrame, dims: str, size_name: str = None, total_name: str = "VOLUME"
) -> pd.DataFrame:
    """
    Takes a pandas dataframe and checks for missing values.
    If a column is numeric and contains missing values, replace them with zeros.
    If a column is categorical and contains missing values, replace them with the column name followed by "_unknown".
    Returns a new pandas dataframe.
    @param df: initial dataset
    @param dims: List of discrete dimensions
    @param size_name: Name of column containing segment sizes
    @param total_name: Name of column that contains totals per segment
    @return: prepared dataset
    """

    new_df = df.copy()  # create a copy of the dataframe to avoid modifying the original

    new_df[total_name] = new_df[total_name].apply(float)

    # replace NaN values in numeric columns with zeros
    if size_name is not None:
        new_df[size_name] = new_df[size_name].apply(float)
        new_df[[size_name, total_name]] = new_df[[size_name, total_name]].fillna(0)
        new_df.loc[new_df[size_name] == 0, total_name] = 0
    else:
        new_df[[total_name]] = new_df[[total_name]].fillna(0)

    # replace NaN values in categorical columns with the column name + "_unknown"
    object_columns = list(new_df[dims].select_dtypes("object").columns)
    new_df[object_columns] = new_df[object_columns].fillna(
        new_df[object_columns].apply(lambda x: x.name + "_unknown")
    )
    if size_name is not None:
        new_df = (
            new_df.groupby(by=dims, observed=True)[[total_name, size_name]]
            .sum()
            .reset_index()
        )
    else:
        new_df = (
            new_df.groupby(by=dims, observed=True)[[total_name]].sum().reset_index()
        )
    new_df[object_columns] = new_df[object_columns].astype(str)
    return new_df



import numpy as np
import pandas as pd
from scipy.sparse import vstack, csr_array, issparse
from scipy.linalg import svd, expm

from wise_pizza.solver import solve_lasso, solve_lp


def find_alpha(
    X,
    y_,
    max_nonzeros=None,
    min_nonzeros=None,
    verbose=0,
    max_iter: int = 100,
    solver: str = "lp",
    use_proj=None,
    constrain_signs=True,
    adding_up_regularizer=False,
):
    """
    Find alpha for optimal number of segments
    @param X: Train data
    @param y_: Target
    @param max_nonzeros: Maximum number of features to find
    @param min_nonzeros: Minimum number of features to find
    @param verbose: If set to a truish value, lots of debug info is printed to console
    @param max_iter: Maximum number of iterations to find segments
    @param solver: If this equals to "lp" uses the LP solver, else uses the (recommended) Lasso solver
    @param use_proj: For using projections
    @param constrain_signs: To constrain signs
    @param adding_up_regularizer: Force the contributions of detected segments to add up to the total difference
    ("other"=0). Experimental, use at your own risk
    @return: Fitted model and nonzeros
    """
    solve = solve_lp if solver == "lp" else solve_lasso
    if use_proj is None:
        use_proj = solve == solve_lp

    min_nonzeros, max_nonzeros = clean_up_min_max(min_nonzeros, max_nonzeros)

    if isinstance(X, pd.DataFrame):
        X = X.values

    y_ = np.array(y_).astype(X.dtype)
    if adding_up_regularizer:
        # add an additional row to the problem, to make the averages add up
        y_ = np.concatenate([y_, y_.sum(keepdims=True) * float(adding_up_regularizer)])
        if issparse(X):
            X = csr_array(vstack([X, X.sum(axis=0) * float(adding_up_regularizer)]))
        else:
            X = np.concatenate(
                [X, X.sum(axis=0, keepdims=True) * float(adding_up_regularizer)]
            )

    sparse_proj = True

    if sparse_proj:
        # Let's collapse the dimension of X and y by projecting both onto X
        # X = U @ S @ Vh so U.shape[0] == X.shape[0] == len(y)
        # Then X.T @ X = Vh.T.@S@S@Vh, and
        # for any matrix M, U.T @ M = Vh.T @ inv(S) @ X.T @ M
        # It's done in this way so the only huge matrix is X,
        # and we never have to create a matrix with one of the dimensions=len(X)

        XtX = (X.T.dot(X)).todense()
        _, s2, Vh = svd(XtX)
        s = np.sqrt(s2)
        eps = 1e-10 * s.max()
        s[s < eps] = eps
        Sinv = np.diag(1 / s)
        H = Sinv.dot(Vh)

        small_mat = H.dot(XtX)
        small_y = H.dot(X.T.dot(y_))
    else:
        # All this is just games on low-dimensional matrices
        # in order to diagnose why Lasso is not invariant to
        # simulateneous rotations of X and y_
        # let's collapse X and y onto the subspace spanned by X
        Xd = X.todense() if issparse(X) else X
        U, _, _ = svd(Xd, full_matrices=False)
        # let's just do a random rotation, closeness to np.eye parametrized by eps
        # eps = 1.0
        # randmat = np.random.randn(len(y_), len(y_))
        # U = expm(eps * (randmat.T - randmat))

        print("Biggest U deviation from eye:", np.max(np.abs(U - np.eye(len(U)))))
        maxerr = np.max(np.abs(U.T.dot(U) - np.eye(len(U))))
        # assert maxerr < 1e-5
        print("Biggest U.T@U deviation from eye:", maxerr)

        # # If we use a degenerate U, it all works
        # U = np.eye(len(U))

        # y: n x 1
        # X: n x m
        # a: m x 1
        #
        # err = y - X @ a
        # loss = err.T @ err + c * sum(abs(a))
        #
        # Xnew = U @ X
        # ynew = U @ y
        # err_new = U @ err
        # loss_new = loss

        small_mat = U.T.dot(Xd)
        small_y = U.T.dot(y_)
    # use for testing
    # assert np.max(np.abs(X.T.dot(X) - small_mat.T.dot(small_mat))) < 1e-2
    # assert (
    #     np.max(np.abs(X.T.dot(y_) - small_mat.T.dot(small_y)))
    #     / np.max(np.abs(small_mat.T.dot(small_y)))
    #     < 1e-5
    # )

    if use_proj:
        mat = small_mat
        y = small_y
    else:
        mat = X
        y = y_

    alpha = 2 * max(y)
    nonzeros = []
    iter = 0

    def print_errors(a: np.ndarray):
        err1 = X.dot(a) - y_
        err2 = np.array(small_mat.dot(a) - small_y)
        yerr = (y_ * y_).sum() - (small_y * small_y).sum()
        sqerr1 = np.sqrt((err1 * err1).sum() - yerr)
        sqerr2 = np.sqrt((err2 * err2).sum())
        d = (err1 * err1).sum() - (err2 * err2).sum() - yerr
        if verbose is not None:
            print("errors", sqerr1, sqerr2)

    if verbose:
        print_errors(np.zeros(X.shape[1]))

    while len(nonzeros) < min_nonzeros:
        alpha /= 2
        reg = solve(
            mat,
            y,
            alpha,
            constrain_signs=constrain_signs,
            verbose=verbose,
            drop_last_row=adding_up_regularizer,
        )
        if verbose:
            print_errors(reg.coef_)
        nonzeros = np.nonzero(reg.coef_)[0]
        iter += 1
        if iter > max_iter:
            break

    min_alpha = alpha
    while len(nonzeros) > max_nonzeros:
        alpha *= 2
        reg = solve(
            mat,
            y,
            alpha,
            constrain_signs=constrain_signs,
            verbose=verbose,
            drop_last_row=adding_up_regularizer,
        )
        if verbose:
            print_errors(reg.coef_)
        nonzeros = np.nonzero(reg.coef_)[0]
        iter += 1
        if iter > max_iter:
            break

    max_alpha = alpha

    # assert iter <= max_iter, "Alpha search failed - max_iter exceeded"
    # TODO: do the second part only on the first set of nonzeros
    alpha_lo = min_alpha
    alpha_hi = max_alpha
    while len(nonzeros) > max_nonzeros or len(nonzeros) < min_nonzeros:
        alpha = (alpha_hi + alpha_lo) / 2
        reg = solve(
            mat,
            y,
            alpha,
            constrain_signs=constrain_signs,
            verbose=verbose,
            drop_last_row=adding_up_regularizer,
        )
        if verbose:
            print_errors(reg.coef_)
        nonzeros = np.nonzero(reg.coef_)[0]
        iter += 1
        if iter > max_iter:
            break
        # assert iter <= max_iter, "Alpha search failed - max_iter exceeded"

        if len(nonzeros) < min_nonzeros:
            alpha_hi = alpha
        elif len(nonzeros) > max_nonzeros:
            alpha_lo = alpha
    if verbose:
        print("alpha=", alpha, nonzeros)
    X_filt = mat[:, nonzeros]
    reg = solve(
        X_filt,
        y,
        1e-3 * alpha,
        constrain_signs=constrain_signs,
        verbose=verbose,
        drop_last_row=adding_up_regularizer,
    )
    return reg, nonzeros


def clean_up_min_max(min_nonzeros: int = None, max_nonzeros: int = None):
    assert min_nonzeros is not None or max_nonzeros is not None
    if max_nonzeros is None:
        if min_nonzeros is None:
            max_nonzeros = 5
            min_nonzeros = 5
        else:
            max_nonzeros = min_nonzeros
    else:
        if min_nonzeros is None:
            min_nonzeros = max_nonzeros

    assert min_nonzeros <= max_nonzeros
    return min_nonzeros, max_nonzeros



from setuptools import find_packages, setup


def _read_requirements_file(path: str):
    with open(path) as f:
        return list(map(
            lambda req: req.strip(),
            f.readlines(),
        ))


with open('README.md') as f:
    long_description = f.read()

setup(
    name="wise-pizza",
    version="0.1.8",
    description="A library to find and visualise the most interesting slices in multidimensional data",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Wise",
    url='https://github.com/transferwise/wise-pizza',
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    install_requires=_read_requirements_file('requirements.txt'),
    extras_require={
        "test": _read_requirements_file('requirements-dev.txt'),
    },
    packages=find_packages(
        include=[
            'wise_pizza',
            'wise_pizza.*'
        ],
        exclude=['tests*'],
    ),
    include_package_data=True,
    keywords='wise-pizza',
)



from dataclasses import dataclass
from typing import List, Optional

import pandas as pd


@dataclass
class SegmentData:
    data: pd.DataFrame
    dimensions: List[str]
    segment_total: str
    segment_size: Optional[str] = None
    segment_std: Optional[str] = None

    def mean(self):
        total = self.data[self.segment_total].sum()
        if self.segment_size is None:
            weight = len(self.data)
        else:
            weight = self.data[self.segment_size].sum()

        return total / weight
